# local imports:
from ..components import TransparentMisc

# standard libraries:
from typing import Optional, Literal, Any
from tkinter import Entry, Event, Misc


class EnhEntry(TransparentMisc, Entry):
    _default_color: Optional[str] = None
    _default_show: Optional[str] = None

    placeholder_color: Optional[str] = None
    placeholder_text: Optional[str] = None

    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            borderwidth: Optional[int] = 6,
            highlightthickness: Optional[int] = 1,
            placeholdercolor: Optional[str] = "#aaaaaa",
            placeholdertext: Optional[str] = "",
            relief: Literal["raised", "sunken", "flat", "ridge", "solid", "groove"] = "flat",
            **standard_options: Any) -> None:
        super().__init__(master, borderwidth=borderwidth, highlightthickness=highlightthickness, relief=relief, **standard_options)
        self.configure(placeholdercolor=placeholdercolor, placeholdertext=placeholdertext)
        self.on_focus_out(event=None)
        self.bind(sequence="<FocusIn>", func=self.on_focus_in, add=True)
        self.bind(sequence="<FocusOut>", func=self.on_focus_out, add=True)

    def on_focus_in(self, event: Optional[Event] = None) -> None:
        if self["foreground"] != self.placeholder_color:
            return None
        self.delete(first="0", last="end")
        Misc.configure(self, foreground=self._default_color, show=self._default_show)

    def on_focus_out(self, event: Optional[Event] = None) -> None:
        if self.get():
            return None
        Misc.configure(self, foreground=self.placeholder_color, show="")
        self.insert(index="0", string=self.placeholder_text)

    def configure(self, **standard_options: Any) -> Any:
        self.on_focus_in(event=None)
        placeholder_color: Optional[str] = standard_options.pop("placeholdercolor", None)
        if isinstance(placeholder_color, str):
            self.placeholder_color = placeholder_color
        placeholder_text: Optional[str] = standard_options.pop("placeholdertext", None)
        if placeholder_text is not None:
            self.placeholder_text = str(placeholder_text)
        result: Any = super().configure(**standard_options)
        self._default_color = self["foreground"]
        self._default_show = self["show"]
        if not self.focus_get():
            self.on_focus_out(event=None)
        return result
    config = configure

    def set_text(self, text: str) -> None:
        self.on_focus_in(event=None)
        self.delete(first="0", last="end")
        self.insert(index="0", string=text)
        if self.focus_get():
            return None
        self.on_focus_out(event=None)

    def cget(self, key: str) -> Any:
        if key == "placeholdercolor":
            return self.placeholder_color
        if key == "placeholdertext":
            return self.placeholder_text
        return super().cget(key)



# local imports:
from ..components import TransparentMisc, ForegroundMisc
from ..utils import FontDescription

# standard libraries:
from tkinter import Event, Label, Misc
from typing import Optional, Any
from tkinter.font import Font


class EnhLabel(ForegroundMisc, TransparentMisc, Label):
    _text: str = ""

    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            background: Optional[str] = "transparent",
            foreground: Optional[str] = "auto",
            padx: Optional[int] = 3,
            text: Optional[str] = "Enhanced Label",
            **standard_options: Any) -> None:
        super().__init__(master, background=background, foreground=foreground, **standard_options)
        self.configure(padx=padx, text=text)
        self.bind(sequence="<Configure>", func=self.on_configure, add=True)

    def on_configure(self, event: Optional[Event] = None) -> None:
        font: FontDescription = self.cget(key="font")
        font = Font(font=font)
        label_width: int = self.winfo_width()
        minimum_width: int = font.measure(text="...")
        if minimum_width > label_width:
            Misc.configure(self, text="...")
            return None
        text: str = self._text
        text_width: int = font.measure(text)
        if text_width > label_width:
            text_width += minimum_width
            while text_width > label_width:
                text = text[:-1]
                text_width = font.measure(text=text + "...")
            text += "..."
        Misc.configure(self, text=text)

    def configure(self, **standard_options: Any) -> Any:
        padx: Optional[int] = standard_options.pop("padx", None)
        if isinstance(padx, int):
            standard_options["padx"] = max(3, padx)
        text: Optional[str] = standard_options.get("text", None)
        if text is not None:
            self._text = str(text)
        return super().configure(**standard_options)
    config = configure

    def cget(self, key: str) -> Any:
        return self._text if key == "text" else super().cget(key)



# local imports:
from .enhanced_widgets import EnhListbox, EnhCanvas, EnhFrame, EnhText
from .enhanced_label import EnhLabel
from .enhanced_image import EnhImage
from .enhanced_window import EnhToplevel, EnhTk
from .enhanced_button import EnhButton
from .enhanced_entry import EnhEntry
from .enhanced_scrollableframe import EnhScrollableFrame



# local imports:
from ..components import ForegroundMisc, HoverMisc
from ..utils import ColorUtils

# standard libraries:
from typing import Optional, Tuple, Any
from tkinter import Button, Misc


class EnhButton(ForegroundMisc, HoverMisc, Button):
    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            background: Optional[str] = "transparent",
            borderwidth: Optional[int] = 0,
            foreground: Optional[str] = "auto",
            hoverbackground: Optional[str] = None,
            text: Optional[str] = "Enhanced Button",
            **standard_options: Any) -> None:
        super().__init__(
            master,
            background=background,
            borderwidth=borderwidth,
            foreground=foreground,
            hoverbackground=hoverbackground,
            text=text,
            **standard_options)
        if hoverbackground is not None:
            return None
        active_background_color: str = self.cget(key="activebackground")
        if not ColorUtils.is_hex_code_valid(hex_code=active_background_color):
            rgb: Tuple[int, int, int] = self.winfo_rgb(color=active_background_color)
            red, green, blue = rgb
            active_background_color = ColorUtils.convert_rgb_to_hex(red, green, blue)
        hover_background_color: str = ColorUtils.adjust_hex_code(
            hex_code=active_background_color,
            brightness_factor=.6)
        self.configure(hoverbackground=hover_background_color)



# local imports:
from . import TransparentMisc, ForegroundMisc
from ..utils import ColorUtils

# standard libraries:
from typing import Generator, Optional, Tuple, Any
from tkinter import Event, Misc


class HoverMisc(TransparentMisc):
    auto_hover_background: bool = False
    background_color_did_update: bool = False
    hover_background_color: Optional[str] = None

    palette_colors: Tuple[str, ...] = ()
    palette_color_amount: int = 12

    cursor_hovered: bool = False
    speed: int = 1
    acceleration_ms: int = 6
    scheduled_update: Optional[str] = None

    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            hoverbackground: Optional[str] = None,
            **standard_options: Any) -> None:
        super().__init__(master, **standard_options)
        self.auto_hover_background = hoverbackground is None
        self.configure(hoverbackground=hoverbackground)
        self.bind(sequence="<<ParentConfigure>>", func=self.update_palette_colors, add=True)
        self.bind(sequence="<Enter>", func=self.on_cursor_interact, add=True)
        self.bind(sequence="<Leave>", func=self.on_cursor_interact, add=True)

    def on_cursor_interact(self, event: Event = None) -> None:
        self.cursor_hovered = int(event.type) == 7
        if self.scheduled_update:
            self.after_cancel(id=self.scheduled_update)
            self.scheduled_update = None
        self.scheduled_update = self.after(ms=self.acceleration_ms, func=self.update_frame)

    def update_frame(self) -> None:
        new_speed: int = self.speed + (1 if self.cursor_hovered else -1)
        new_speed = min(new_speed, self.palette_color_amount)
        new_speed = max(new_speed, 1)
        if new_speed == self.speed:
            return None
        self.speed = new_speed
        try:
            background_color: str = self.palette_colors[new_speed - 1]
            Misc.configure(self, background=background_color)
            self.event_generate(sequence="<<BackgroundConfigure>>")
            self.update_children()
        except IndexError:
            pass
        self.scheduled_update = self.after(ms=self.acceleration_ms, func=self.update_frame)

    def update_palette_colors(self, event: Optional[Event] = None) -> None:
        background_color: str = Misc.cget(self, key="background") \
            if not self.palette_colors or event is not None \
            else self.palette_colors[0]
        if background_color == "transparent":
            background_color = Misc.cget(self.master, key="background") or Misc.cget(self, key="background")
        if not ColorUtils.is_hex_code_valid(hex_code=background_color):
            rgb: Tuple[int, int, int] = self.winfo_rgb(color=background_color)
            red, green, blue = rgb
            background_color = ColorUtils.convert_rgb_to_hex(red, green, blue)
        palette_colors: Generator[str, None, str] = ColorUtils.generate_palette_colors(
            start_hex_code=background_color,
            end_hex_code=self.hover_background_color,
            color_amount=self.palette_color_amount)
        self.palette_colors = tuple(palette_colors)
        try:
            background_color: str = self.palette_colors[self.speed - 1]
            Misc.configure(self, background=background_color)
            self.event_generate(sequence="<<BackgroundConfigure>>")
            self.update_children()
        except IndexError:
            pass

    def configure(self, **standard_options: Any) -> None:
        background_color: Optional[str] = standard_options.get("background", None)
        background_color = standard_options.get("bg", background_color)
        hover_background_color: Optional[str] = standard_options.pop("hoverbackground", None)
        if background_color is not None and self.auto_hover_background:
            hex_code: str = background_color or Misc.cget(self, key="background")
            if hex_code == "transparent":
                hex_code = Misc.cget(self.master, key="background") or Misc.cget(self, key="background")
            if not ColorUtils.is_hex_code_valid(hex_code):
                rgb: Tuple[int, int, int] = self.winfo_rgb(color=hex_code)
                red, green, blue = rgb
                hex_code = ColorUtils.convert_rgb_to_hex(red, green, blue)
            hover_background_color = hover_background_color or ColorUtils.adjust_hex_code(
                hex_code, brightness_factor=3.)
        if hover_background_color is not None or self.hover_background_color is None:
            if hover_background_color is None:
                hex_code: str = background_color or Misc.cget(self, key="background")
                if hex_code == "transparent":
                    hex_code = Misc.cget(self.master, key="background") or Misc.cget(self, key="background")
                if not ColorUtils.is_hex_code_valid(hex_code):
                    rgb: Tuple[int, int, int] = self.winfo_rgb(color=hex_code)
                    red, green, blue = rgb
                    hex_code = ColorUtils.convert_rgb_to_hex(red, green, blue)
                hover_background_color = ColorUtils.adjust_hex_code(hex_code, brightness_factor=3.)
            self.hover_background_color = hover_background_color
        result: Any = super().configure(**standard_options)
        if background_color is not None or hover_background_color is not None:
            self.update_palette_colors(event=None)
        return result
    config = configure

    def cget(self, key: str) -> Any:
        return self.hover_background_color if key == "hoverbackground" else super().cget(key)



# local imports:
from ..utils import ColorUtils

# standard libraries:
from tkinter import Widget, Event, Misc
from typing import Optional, Any


class ForegroundMisc(Misc):
    is_auto_foreground: bool = False

    def __init__(self, master: Optional[Misc] = None, *, foreground: Optional[str] = "auto", **standard_options: Any) -> None:
        assert isinstance(self, Widget), "This instance must be a Widget."
        foreground_color: Optional[str] = standard_options.pop("fg", foreground)
        super().__init__(master, **standard_options)
        self.configure(foreground=foreground_color)
        self.bind(sequence="<<BackgroundConfigure>>", func=self.on_background_configure, add=True)

    def on_background_configure(self, event: Optional[Event] = None) -> None:
        if not self.is_auto_foreground:
            return None
        hex_code: str = self["background"]
        if not ColorUtils.is_hex_code_valid(hex_code):
            red, green, blue = self.winfo_rgb(color=hex_code)
            hex_code = ColorUtils.convert_rgb_to_hex(red, green, blue)
        foreground_color: str = "#ffffff" if ColorUtils.is_hex_code_darker(hex_code) else "#000000"
        Misc.configure(self, foreground=foreground_color)

    def configure(self, **standard_options: Any) -> Any:
        foreground_color: Optional[str] = standard_options.pop("foreground", None)
        foreground_color = standard_options.pop("fg", foreground_color)
        if foreground_color is not None:
            self.is_auto_foreground = isinstance(foreground_color, str) and foreground_color == "auto"
            if not self.is_auto_foreground:
                standard_options["foreground"] = foreground_color
        result: Any = super().configure(**standard_options)
        if self.is_auto_foreground:
            self.on_background_configure(event=None)
        return result
    config = configure

    def cget(self, key: str) -> Any:
        return "auto" if key in ("fg", "foreground") and self.is_auto_foreground else super().cget(key)



# standard libraries:
from setuptools import setup, find_packages
from typing import Dict, List, Any

# TODO: pip install twine wheel
# TODO: python setup.py sdist bdist_wheel
# TODO: twine upload dist/*

# setup:
application_packages: List[str] = find_packages()
application_settings: Dict[str, Any] = {
    "name": "TkEnhanced",
    "version": "1.1.1",
    "description": "Enhanced Tkinter widgets for a modern look and additional functionality.",
    "long_description": "This package includes enhanced widgets for Tkinter with a modern appearance and additional features. It aims to improve the aesthetics and functionality of Tkinter applications.",
    "long_description_content_type": "text/markdown",
    "author": "Bunnitown",
    "packages": application_packages,
    "install_requires": ["pillow>=10.4.0", "numpy>=2.0.1"],
    "classifiers": [
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]}
setup(**application_settings)



# local imports:
from ..utils import WindowUtils, PathDescription

# standard libraries:
from tkinter import Toplevel, Misc, Tk
from typing import Optional, Any
from os.path import exists


class Window(WindowUtils):
    def configure_window(self, *, title: str, width: int, height: int, icon_path: PathDescription = None) -> None:
        self.wm_title(string=title)
        if icon_path and exists(path=icon_path):
            self.wm_iconbitmap(default=icon_path)
        self.center_window(width=width, height=height)

    def update_children(self) -> None:
        for child_widget in self.winfo_children():
            child_widget.event_generate(sequence="<<ParentConfigure>>")

    def configure(self, **standard_options: Any) -> Any:
        background_color: Optional[str] = standard_options.get("background", None)
        background_color = standard_options.get("bg", background_color)
        result: Any = super().configure(**standard_options)
        if background_color is not None:
            self.update_children()
        return result
    config = configure


class EnhTk(Window, Tk):
    def __init__(
            self,
            *,
            title: Optional[str] = "Enhanced Tk",
            width: Optional[int] = 800,
            height: Optional[int] = 600,
            icon_path: PathDescription = None,
            **standard_options: Any) -> None:
        standard_options["className"] = standard_options.pop("className", "enhanced_tk")
        super().__init__(**standard_options)
        self.configure_window(title=title, width=width, height=height, icon_path=icon_path)


class EnhToplevel(Window, Toplevel):
    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            title: Optional[str] = "Enhanced Toplevel",
            width: Optional[int] = 800,
            height: Optional[int] = 600,
            icon_path: PathDescription = None,
            **standard_options: Any) -> None:
        super().__init__(master, **standard_options)
        self.configure_window(title=title, width=width, height=height, icon_path=icon_path)
